/*
 Hack program to process APOD image of 7-Dec-2004 (based on code
 used to process scans of old maps).

 Reads images in 24-bit raw format (, the dimensions are hardcoded below), displays
 the image and processes it in various ways.

 Never intended for public release; it was expected this code would be thrown away
 after processing the APOD images. However requests have been made for the release
 of this code, and it is therefore released as GPL code (www.gnu.org).

 No warranties whatsoever.

 For additional information see http://hazeii.net/apod

 Copyright (c) N Newell 2004-2005 unless otherwise noted.

 Initial release version: 0.0.1 19-Dec-2004
 Various updates and enhancements: 0.0.2 2-Jan-2005

*/

#define BPP 3	/* Use 1 for 8-bit internal processing, 3 for 24-bit */

/* Filename and dimensions of image in disk file */
#if 0
 #define FILENAME "map3308x2450.raw" /**/
 #define IMGX 3308	/* Test image width */
 #define IMGY 2450	/* Test image height */
 #define COLOUR 1
#elif 0
 #define FILENAME "/pub/trail1024x200.raw" /**/
 #define IMGX 1024	/* Test image width */
 #define IMGY 200	/* Test image height */
 #define COLOUR 0 /* Greyscale image */
#elif 0
 #define FILENAME "/pub/greydiff1358x960-33d6.raw" /* MAIN IMAGE USED ON APOD*/
 #define IMGX 1358	/* Test image width */
 #define IMGY 960	/* Test image height */
 #define COLOUR 0 /* Greyscale image */
#elif 0
 #define FILENAME "mout.raw" /* "diffbam.raw" /**/
 #define IMGX 2272	/* Test image width */
 #define IMGY 1704	/* Test image height */
 #define COLOUR 0 /* Greyscale image */
 #define XSTART 436
 #define YSTART 608
#elif 1
 #define FILENAME "trail1250x200.raw" /**/
 #define IMGX 1250	/* Test image width */
 #define IMGY 200	/* Test image height */
 #define COLOUR 0 /* Greyscale image */
 #define XSTART 0
 #define YSTART 0 // 20
#elif 1
 #define FILENAME "dr3600x2700.raw" /**/
/* #define FILENAME "dr3600x2700sim.raw" /**/
/* #define FILENAME "dr3600x2700bugsim.raw" /**/
 #define IMGX 3600	/* Test image width */
 #define IMGY 2700	/* Test image height */
 #define XSTART 227
 #define YSTART 466
// #define XSTART 227 /* For simulated image */
// #define YSTART 443
 #define COLOUR 0 /* Greyscale image */
#else
 #define FILENAME "/pub/greydiff2184x1544-33d8.raw" /* */
 #define IMGX 2184	/* Test image width */
 #define IMGY 1544	/* Test image height */
 #define COLOUR 0 /* Greyscale image */
#endif

/* 76,131 for main APOD image */

#ifndef XSTART
 #define XSTART  0//76 // 1600 // 500
 #define YSTART  0//131 // 1438 // 1038
#endif

/* Requested display window size (note we may get something slightly different) */

#define WXSIZE  1250 // 1270 // 250 // (1280-20)			/* Required X dimension */
#define WYSIZE  200 // 200 // 250 // (1024-50)			/* Required Y dimension */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <math.h>
#include <time.h>

//#include "map.h"

static struct {
	int width;
	int height;
	unsigned char *dsptr;
	unsigned char *backptr;
	} gd;

#define CTL(x) (x&31)

/* Local Prototypes */

void redraw();

#define XP printf

void ttyout(char c);
void putstr(char *p);

static void gfxon(void);

static int map_read(unsigned x,unsigned y,unsigned xsize,unsigned ysize,unsigned buffwidth,unsigned buffheight,unsigned char *buff);

unsigned char *imgbuff;

int cortest(double factor,int xsize,int ysize,unsigned char *pdata);
int bugtest(int xsize,int ysize,int xlimit,unsigned char *pdata);

/************************* main *************************/

static unsigned run,paused;

main(int argc,char *argv[])
{
int i,j,x,y,mangle,dobug,doautocor,xlimit,xsize,ysize;
unsigned char *p,*fname,buff[200];
unsigned recolour[3]={0x60,0xb0,0xf4};	/* Light blue grid lines */
//unsigned recolour[3]={0xf7,0x70,0xa0}; /* Red footpath/bridleway lines */
double acfactor;

	if (argc > 1)
		fname=argv[--argc];
	else
		fname = FILENAME;

/* Start graphics and create display window */

	gfxon();

	xsize = gd.width;
	ysize = gd.height;
	printf("Window xsize, ysize %u %u\n",xsize,ysize);

/* Allocate memory for image buffer */

	if ((imgbuff=malloc(xsize*ysize*3))==NULL) {
		XP("Unable to allocate image data buffer\n");
		exit(1);
		}

/* Open the file */

	if (map_open(fname)==0) {
		exit(1);
		}

/* Main loop */

	run = 1;

X_color(1);

dobug=doautocor=0;
acfactor=2.5;

/* Set initial image position */

		x = XSTART;
		y = YSTART;

/* Main loop */

	mangle = 0;

	while (run) {

		set_recolour(recolour[0],recolour[1],recolour[2]);

		map_read(x,y,IMGX,IMGY,xsize,ysize,imgbuff);

		xlimit = xsize-35;	/* Horizontal distance to process over */

		if (doautocor) {
			cortest(acfactor,xsize,ysize,imgbuff);
			xlimit = xsize/acfactor;
			}

		if (dobug)
			bugtest(xsize,ysize,xlimit,imgbuff);

		test(xsize,ysize,imgbuff,mangle);

		sprintf(buff,"Image position = %u,%u",x,y);
		X_string(buff,150);
		paint();

		while ((i=getkey()) <= 0)
			;

		while (i >= 0) {

			if (i == 0x1b) {
				run = 0;
				break;
				}

//XP("Key 0x%x  CTLB = 0x%x\n",i,CTL('B'));
#define SPIX 1
			if (i==CTL('B'))
				x=x>SPIX?x-SPIX:0;
			else if (i==CTL('F'))
				x=x+SPIX<IMGX-xsize?x+SPIX:IMGX-xsize;
			else if (i==CTL('P'))
				y=y>SPIX?y-SPIX:0;
			else if (i==CTL('N'))
				y=y+SPIX<IMGY-ysize?y+SPIX:IMGY-ysize;

			if (i=='+')
				;
			else if (i=='-')
				;
			if (i=='>') {
				if (acfactor >= 1.1)
					acfactor -= 0.1;
				XP("AC Factor %0.2lf\n",acfactor);
				}
			else if (i=='<') {
				acfactor += 0.1;
				XP("AC Factor %0.2lf\n",acfactor);
				}
			else if (i=='t')
				;
			else if (i=='#')
				dobug=!dobug;
			else if (i=='c')
				doautocor=!doautocor;
			else if (i=='d')
				redraw();

			if (i=='z')	/* Reset all image transform options? */
				mangle=0;
			else if (i=='m')	/* Greyscale */
				mangle^=1;
			else if (i=='h')	/* Horizontal edge */
				mangle^=2;
			else if (i=='v')	/* Vertical edge */
				mangle^=4;
#if 0
			else if (i=='x')	/* Convolution filter */
				mangle^=8;
#endif
			else if (i=='f')	/* False colour */
				mangle^=16;

			else if (i=='*') {
				X_locator(&i,&j);
				if (i>=0 && i < xsize && j >= 0 && j < ysize) {
					printf("Mouse at %u,%u",i,j);
					p = imgbuff + (i+(ysize-1-j)*xsize)*3;
					XP("  R: 0x%02x ",*p++);
					XP("G: 0x%02x ",*p++);
					XP("B: 0x%02x ",*p++);
					XP("\n");
					}
				}

/* Modify target colourRGB values */

			if (i=='r'&&recolour[0]>0)
				recolour[0]--;
			else if (i=='R'&&recolour[0]<255)
				recolour[0]++;

			if (i=='g'&&recolour[1]>0)
				recolour[1]--;
			else if (i=='G'&&recolour[1]<255)
				recolour[1]++;

			if (i=='b'&&recolour[2]>0)
				recolour[2]--;
			else if (i=='B'&&recolour[2]<255)
				recolour[2]++;

/* Check for another key */

			i = getkey();
			}
		}

/* Program terminating - shut down graphics and exit */

	gfxoff();

	XP("\nProgram terminated\n");
}

#if 0
graphmode(int mode)
{
	return gfxmode = mode;
}
getgraphmode(e)
{
	return gfxmode;
}
#endif
/**************************** key handling *********************************/

static unsigned char keybuff[50];
static unsigned int keyput;
static unsigned int keyget;

putkey(unsigned char c)
{
	keybuff[keyput] = c;

	if (++keyput >= sizeof(keybuff))
		keyput = 0;

	if (keyput == keyget)
		if (++keyget >= sizeof(keybuff))
			keyget = 0;
}

getkey()
{
unsigned char c;

	if (os_scan() < 0)
		run = 0;

	if (keyput == keyget) {
		sched_yield();
		return -1;
		}

	c = keybuff[keyget];

	if (++keyget >= sizeof(keybuff))
		keyget = 0;

	return c;
}


/**************** os_scan *****************/

/* Routine called from main() application */

os_scan()
{
	X_check_keypress();

	return 0;
}

/******************** paint ************************/
paint()
{
int i,x,y;
unsigned char buff[256];
unsigned char *p,*s,points[1024];

//	gfxcolumn(points,y);

	statusbar(1);

	gfxpaint();

	X_update(1);
	X_clear();

	return 1;
}
/************************* blit *******************************/

blit(unsigned x,unsigned y,unsigned xsize,unsigned ysize,unsigned char *ptr)
{
unsigned i;
unsigned char *p;

/* Validate/clip X */

	if (x >= gd.width)
		return 0;

	if (x+xsize > gd.width)
		xsize = gd.width - x;

/* Validate/clip Y */

	if (y >= gd.height)
		return 0;

	if (y+ysize > gd.height)
		ysize = gd.height - y;

/* Blit data into bitmap buffer */

	p = gd.dsptr + y*gd.width + x;

/*XP("X,Y %d %d Size %d %d  Index %d  %p/%p %p\n",x,y,xsize,ysize,y*gd.width+x,gd.dsptr,p,ptr); /**/

	for (i=0;i<ysize;i++) {
		memcpy(p,ptr,xsize); /**/
		p += gd.width;
		ptr += xsize;
		}

	return 1;
}
/********************** vline ************************/

vline(unsigned sx,unsigned sy,unsigned ylen,unsigned colour)
{
int i;
unsigned char *p;

/* Validate/clip X */

	if (sx >= gd.width)
		return 0;

/* Validate/clip Y */

	if (sy >= gd.height)
		return 0;

	if (sy+ylen > gd.height)
		ylen = gd.height - sy;

/* Draw the line */

	p = gd.dsptr + sy*gd.width + sx;

	for (i=0;i<ylen;i++) {
		*p = colour; /**/
		p+= gd.width;
		}

	return 1;
}
/********************** hline ************************/

hline(unsigned sx,unsigned sy,unsigned xlen,unsigned colour)
{
int i;
unsigned char *p;

/* Validate/clip X */

	if (sx >= gd.width)
		return 0;

	if (sx+xlen > gd.width)
		xlen = gd.width-sx;

/*XP("SX %u  gd.width %u  sx+xlen %u\n",sx,gd.width,sx+xlen);  /**/

/* Validate Y */

	if (sy >= gd.height)
		return 0;

/* Draw the line */

	p = gd.dsptr + sy*gd.width + sx;

	memset(p,colour,xlen); /**/

	return 1;
}
/************************ plot ***********************/

plot(unsigned sx,unsigned sy,unsigned colour)
{

/* Validate pixel position */

	if (sx >= gd.width || sy >= gd.height)
		return 0;

/* Plot the point */

	*(gd.dsptr + sy*gd.width + sx) = colour;

	return 1;
}
/************************** getcharsize ***************************/

getcharsize(unsigned int *xp, unsigned int *yp)
{
	*xp = 8;
	*yp = 16;
}
/************************* clear_text ************************/
clear_text()
{
//	memset(screen,' ',sizeof(screen));
	return 1;
}
/*************************** clear_screen *********************/
clear_screen()
{
	if (gd.dsptr != NULL)
		memset(gd.dsptr,0,gd.width*gd.height);

	return 1;
}
/********************** putstr ***********************/
void putstr(char * str)
{
	while (*str)
		ttyout(*str++);
}
/************************* ttyout **************************/
void ttyout(char c)
{
	XP("%c",c);
}
/***************************** XP **********************************/

void __XP(char *fmt,...)
{
int i;
va_list argptr;
char c,*p,buff[500];

/* Create string in buffer */

	va_start(argptr,fmt);	/* Initialize variable argument list */

	vsprintf(buff,fmt,argptr);	/* Print string with variable arg. list */

	printf(buff);

	va_end(argptr);		/* Clean up argument list */

	return;
}
/* ============================ */

static void gfxon()
{
int wxsize,wysize;

	wxsize = WXSIZE;
	wysize = WYSIZE;
	X_init(&wxsize,&wysize);
	X_clear();
	X_backbuf();

	gd.width = wxsize;
	gd.height = wysize;
	gd.dsptr = malloc(wxsize*wysize*BPP);

XP("Graphics on - display size %ux%u\n",wxsize,wysize);

}
gfxoff()
{
	X_exit();
}
/* ================================================== */

static FILE *fp;

int map_open(unsigned char *fname)
{
	if (fp != NULL) {
		fclose(fp);
		fp = NULL;
		}

	if ((fp = fopen(fname,"rb")) == NULL) {
		XP("Cannot open file %s\n",fname);
		return 0;
		}

XP("Opened image file %s\n",fname);

	return 1;

}

/************************* map_read *************************/

/* Read routine used when internal processing is 24 bits per pixel */

static int map_read(unsigned x,unsigned y,unsigned xsize,unsigned ysize,unsigned width,unsigned height,unsigned char *buff)
{
int i,j,k,buffsize,bpp;
unsigned index,maxndx;
unsigned char *p,*s,*t;

#if COLOUR
	bpp=3;
#else
	bpp=1;
#endif

	maxndx = xsize*ysize*bpp;		/* Numnber of bytes expected in file */

	index = (y*xsize+x)*bpp;		/* Start read point */

XP("Image offset: X %u Y %u\n",x,y);
/*XP("Window Height: %u  File y dimension: %u\n",height,ysize); /**/

	if (index >= maxndx)
		return 0;

	p = buff;
	fseek(fp,index,SEEK_SET);

	for (i=0;i<height;i++) {

		if ((j = fread(p,1,width*bpp,fp))<=0)
			break;

		if (bpp==1 && BPP==3) {	/* One byte per pixel in file, 24-bit internal? */

			s=p+width;	/* Replicate each byte 3 times */
			t=p+width*BPP;

			for (k=0;k<width;k++) {
				*--t=*--s;
				*--t=*s;
				*--t=*s;
				}
			}

		p+=width*BPP;

		index += xsize*bpp;

		fseek(fp,index,SEEK_SET);
		}

#if 0 /* Indexed palette display */
if (i>height-30) {
 for (j=0;j<width;j++)
  *(p+j)=j;
 }
#endif


//XP("Final index %u\n",index);

	return 1;
}

int map_close()
{
	if (fp != NULL)
		fclose(fp);
	fp = NULL;

	return 1;
}
