/* 
 Read colour image, draw on it, and write it out

 The drawing information is contained in text files, contained raw (first
 field) and smoothed (second field) Y values for each X.
*/

/*#define MONO 1	/* Make true if processing mono (8bits/pixel) images */

#define USESMOOTHED 1 /* Make true to use the smoothed data field */

#if MONO
 #define F1 "diffbam.raw"
 #define F2 "mout.raw"
 #define BPP 1
#else
 #define F1 "cmain2272x1704.raw"
 #define F2 "cout.raw"
 #define BPP 3
#endif

#define XSIZE 2272	/* Test image width */
#define YSIZE 1704	/* Test image height */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <math.h>
#include <time.h>

#define XP printf

#ifndef M_PI
 #define M_PI 3.14159265
#endif

int load_lsline(char *fname,double out[][2]);
int remap(double dx,double dy,unsigned *px,unsigned *py);

/************************* main *************************/

main()
{
int i,j,x,y;
unsigned mono;
unsigned char *p1,*p2,*p3,*p4,*pc;
unsigned char *p,*q,*s,*t;
unsigned char buff[200];
FILE *fp;

/* Allocate memory for image buffer */

	if ((pc=malloc(XSIZE*YSIZE*BPP))==NULL) {
		XP("Unable to allocate image data buffer\n");
		exit(1);
		}

/* Load image(s) */
#if 1
	if ((fp = fopen(F1,"rb")) == NULL) {
		XP("Cannot open file %s\n",F1);
		return 0;
		}

	XP("Read %d bytes\n",fread(pc,1,XSIZE*YSIZE*BPP,fp));

	fclose(fp);
#endif
/* Add simulated test streak to main image for control purposes */

	simstreak(pc);	/**/

/* Draw best-fit lines */

/*	lslines(pc); /**/

/* Write out modified image */

	write_image(F2,pc);

	return 0;
}
/************************* simstreak *************************/

/*
 Adds a simulated streak parallel to the actual mystery streak (for use as a control)

 NOTE: ensure this code tracks the code in imdiff.c 

*/

/************************* simstreak *************************/

/*
  Adds a simulated streak parallel to the actual mystery streak (for use as a control)

  NOTE: This code is a copy of the code in imdiff.c, and should be kept identical.

*/

#define XSIMO 160
#define YSIMO 345	/* Originally 340 */

#define ALPHA (251.0/255.0)	/* Streak density: use 251.0 for reasonable fit to image */

#define TOFF (25)	/* Originally zero, but trail hit horizon */

int simstreak(unsigned char *imgptr)
{
int i,j,k,m,len,x,y,yb4,min,max;
unsigned char *s;
double z,v,avg,scale,sina,cosa;

#define SIMANGLE (33.6/180.0*3.14159265)

#define SIMWIDTH 25 // 30 /* Cross-sectional width of the fake track */

	sina=sin(SIMANGLE);
	cosa=cos(SIMANGLE);

	len = SIMWIDTH/cos(SIMANGLE+3.14159/2);
	len=abs(len);

	yb4=0;
	for (i=0;i<1250;i++) {
		x = (double) (i-TOFF) * cosa + XSIMO;
		y = (double) (i-TOFF) * sina + YSIMO;

		if (y != yb4) {
			s = imgptr+(y*XSIZE+x)*BPP;
			for (j=0;j<len;j++) {
				z = (double) (2*j-len)/(double)len;
				z*=3.14159;
				z = -cos(z)/2.0+0.5;	/* Range 0 to 1.0 */
				z = z*z; /* Squaring gives a better fit to real trail */

				for (m=0;m<BPP;m++) {
					k = *s;	/* Get pixel value (0 to 255) */
					k = (k<<8)+(rand()&255);	/* Scale up by 256, add noise */
					v = (double)k;
					v = v*ALPHA + (z*v)*(1.0-ALPHA);
					k = v;
					k>>=8;
					*s++=k;
					}
				}
			}
		yb4=y;
		}

	return 1;
}
/************************* lslines *************************/

#define STRETCH 4.0	/* Amount to stretch trail deviations by */

#define XLOCUT 75	/* Image X to start drawing at */
#define XHICUT 1150	/* Image X to stop drawing at */

int lslines(unsigned char *imgptr)
{
unsigned i,n,x,y,ax,ay;
unsigned int slen,alen;
unsigned char *s;
double actual[2000][2],sim[2000][2];
double sum,avg;

/* Plot the 'actual' line */

	alen = load_lsline("true.lst",actual);

/* Compute average position (along the definite part of the trail) */

	n = 0.0;
	sum = 0.0;

	for (i=0;i<alen;i++) {

		if (i>=250 && i<1150) {	/* Compute average over definite part of trail */
			sum+=actual[i][1];
			n++;
			}
		}

	avg = sum/(double) n;
	XP("Actual: Entries %u Average %.3lf (over %u entries)\n",alen,avg,n);

/* Now plot the actual trail position on the image */

	for (i=0;i<alen;i++) {

		actual[i][1] = (actual[i][1]-avg)*STRETCH + actual[i][1];  

		remap(actual[i][0],actual[i][1],&x,&y);

		if (x>=XSIZE || y >= YSIZE)
			XP("'Actual' line: Off-image point %d,%d!\n",x,y);
		else if (x < XLOCUT || x > XHICUT)
			; 
		else {
#if 1 /* Plot line showing the average */
remap(actual[i][0],avg,&ax,&ay);
s = imgptr+(ay*XSIZE+ax)*BPP;
*s++ = 255;*s++ = 255;*s++ = 0;
*s++ = 255;*s++ = 255;*s++ = 0;
#endif
			s = imgptr+(y*XSIZE+x)*BPP;
			*s++ = 0;
#if BPP == 3
			*s++ = 0;
			*s++ = 0;
#endif
*s++=0;*s++=0;*s++=0;
			}
		}

/* Plot the 'simulated' line */
#if 1
	slen = load_lsline("sim.lst",sim);

/* Compute average position */

	n = 0.0;
	sum = 0.0;

	for (i=0;i<alen;i++) {

		if (i>=250 && i<1150) {
			sum+=sim[i][1];
			n++;
			}
		}

	avg = sum/(double) n;
	XP("Sim: Entries %u Average %.3lf (over %u entries)\n",alen,avg,n);

	for (i=0;i<alen;i++) {
		x = sim[i][0];
		y = sim[i][1];

		sim[i][1] = (sim[i][1]-avg)*STRETCH + sim[i][1];  
		remap(sim[i][0],sim[i][1],&x,&y);

		if (x>=XSIZE || y >= YSIZE)
			XP("'Actual' line: Off-image point %d,%d!\n",x,y);
		else if (x < XLOCUT || x > XHICUT)
			; 
		else {
#if 1 /* Plot line showing the average */
remap(sim[i][0],avg,&ax,&ay);
s = imgptr+(ay*XSIZE+ax)*BPP;
*s++ = 255;*s++ = 255;*s++ = 0;
*s++ = 255;*s++ = 255;*s++ = 0;
#endif
			s = imgptr+(y*XSIZE+x)*BPP;
			*s++ = 0;
#if BPP == 3
			*s++ = 0;
			*s++ = 0;
#endif
*s++=0;*s++=0;*s++=0;
			}
		}

#endif
	return 1;
}

/************************* load_lsline *************************/

int load_lsline(char *fname,double out[][2])
{
unsigned i,x,y,ndx,line;
char buff[500];
double dx,dy;
FILE *fp;


	if ((fp = fopen(fname,"r")) == NULL) {
		XP("Can't access source file [%s]",fname);
		return 0;
		}

	ndx = 0;
	line = 1;

	while (fgets(buff,sizeof(buff),fp) > 0) {

		if (sscanf(buff,"%u %lf",&y,&dy) < 2)
			XP("WARNING - File %s Line %u: bad or empty\n",fname,line);
		else {
			dx=(double)ndx;
			out[ndx][0]=dx;
#if USESMOOTHED
			out[ndx][1]=dy;
#else
			out[ndx][1]=y;
#endif
			ndx++;
			}
		line++;
		}

	fclose(fp);

	XP("File %s - read %u line(s)\n",fname,ndx);

	return ndx;
}
/************************* remap *************************/

#define TSTARTX 170
#define TSTARTY 275 
#define ANGLE (33.6/180.0*M_PI)

#define OXSIZE 1250
#define OYSIZE 200

#define XPRE 150		/* OYSIZE*tan(M_PI/4-ANGLE)/2 */
#define XPOST 150

#define OXTOTAL (XPRE+OXSIZE+XPOST)

/*
 Given a pixel position [dx,dy] remap to the appropriate pixel position in
 the original image (this undoes the original mapping in imdiff.c, so if the
 code or parameters are changed the same change must be made here).
*/

int remap(double dx,double dy,unsigned *px,unsigned *py)
{
int i,ix,iy;
double x1,x2,odx,ody,tana,z;

/* First we need to undo the X shift */

//dx=28;
//dy=102.5;
	tana = -1.0/tan(ANGLE);		/* Tangent of the trail cross-section */
	x1 = -OYSIZE/tana;			/* Horizontal pixels crossed prior to shear */
	tana = (OYSIZE+x1*tan(ANGLE))/x1; /* Tangent of line post-shear (Y+Ydelta)/X1 */
	x2 = OYSIZE/tana;			/* New horizontal length */

	odx = dx - (double) dy/OYSIZE * x2;	/* Subtract the shift originally applied */

//XP("Undo X: %.2lf,%.2lf => %.2lf,%.2lf\n",dx,dy,odx,ody);

/* Now undo the Y shift */

	tana = tan(ANGLE);

//XP("Undo Y: %.2lf,%.2lf => %.2lf,%.2lf\n",odx,ody,odx,ody+odx*tana);

	ody=dy+odx*tana;

/* Translate back to main image coords */

	odx += (TSTARTX-XPRE);
	ody += (TSTARTY-OYSIZE/2);

//XP("Image coords: %.2lf, %.2lf\n",odx,ody);

/* Validate */

	if (odx < 0 || ody < 0 || odx >= XSIZE || ody >= YSIZE) {
/*		XP("Ooops...off-screen value in remap() - dx %.2lf dy %.2lf => %.2lf %.2lf\n",dx,dy,odx,ody); /**/
		odx = ody = 0.0;
		}

/* Return the results */

	*px = odx;
	*py = ody;

	return 1;
}
/************************* write_image *************************/

int write_image(char *name,char *p)
{
unsigned n;
FILE *fp;

	if ((fp = fopen(name,"wb")) == NULL) {
		XP("Write Image error: Can't open %s\n",name);
		exit(1);
		}

	n = XSIZE*YSIZE*BPP;

	if (fwrite(p,1,n,fp) != n)
		XP("Error writing file %s\n",name);

	fclose(fp);

	XP("File \"%s\" written.\n",name);

	return 1;

}
